from collections import Counter
from clarify_python.helper import get_embedded_items, get_link_href

MAX_METADATA_STRING_LEN = 2000


def default_to_empty_string(val):
    return val if val is not None else ''


class ClarifyBrightcoveBridge:

    def __init__(self, clarify_client, bc_client):
        self.clarify_client = clarify_client
        self.bc_client = bc_client
        self.sync_stats = Counter(created=0, updated=0, deleted=0)
        self.dry_run = False

    def log(self, output_str):
        print(output_str)

    def log_sync_stats(self):
        print('\nBundle stats:')
        print('  created: {0}'.format(self.sync_stats['created']))
        print('  updated: {0}'.format(self.sync_stats['updated']))
        print('  deleted: {0}'.format(self.sync_stats['deleted']))
        print('  total:   {0}'.format(self.sync_stats['count']))

    def _load_bundle_map(self):
        '''
        Return a map of all bundles in the Clarify app that have an external_id set for them.
        The bundles with external_ids set are assumed to be the ones we have inserted from Brightcove.
        The external_id contains the Brightcove video id.
        '''
        bundle_map = {}
        next_href = None
        has_next = True

        while has_next:
            bundles = self.clarify_client.get_bundle_list(href=next_href, embed_items=True)
            items = get_embedded_items(bundles)
            for item in items:
                bc_video_id = item.get('external_id')
                if bc_video_id is not None and len(bc_video_id) > 0:
                    bundle_map[bc_video_id] = item
            next_href = get_link_href(bundles, 'next')
            if next_href is None:
                has_next = False
        return bundle_map

    def _metadata_from_video(self, video):
        '''Generate the searchable metadata that we'll store in the bundle for the video'''
        long_desc = video['long_description']
        if long_desc is not None:
            long_desc = long_desc[:MAX_METADATA_STRING_LEN]

        tags = video.get('tags')

        metadata = {
            'name': default_to_empty_string(video.get('name')),
            'description': default_to_empty_string(video.get('description')),
            'long_description': default_to_empty_string(long_desc),
            'tags': tags if tags is not None else [],
            'updated_at': video.get('updated_at'),
            'created_at': video.get('created_at'),
            'state': video.get('state')
        }
        return metadata

    def _src_media_url_for_video(self, video):
        '''Get the url for the video media that we can send to Clarify'''
        src_url = None
        best_height = 0
        best_source = None

        # TODO: This assumes we have ingested videos. For remote videos, check if the remote flag is True
        # and if so, use the src url from the Asset endpoint.
        video_sources = self.bc_client.get_video_sources(video['id'])
        # Look for codec H264 with good resolution
        for source in video_sources:
            height = source.get('height', 0)
            codec = source.get('codec')
            if source.get('src') and codec and codec.upper() == 'H264' and height <= 1080 and height > best_height:
                best_source = source

        if best_source is not None:
            src_url = best_source['src']
        return src_url

    def _create_bundle_for_video(self, video):

        media_url = self._src_media_url_for_video(video)
        if not media_url:
            self.log('SKIPPING: No suitable video src url for video {0} {1}'.format(video['id'], video['name']))
            return

        self.log('Creating bundle for video {0} {1}'.format(video['id'], video['name']))

        if not self.dry_run:
            external_id = video['id']
            name = video.get('original_filename')
            metadata = self._metadata_from_video(video)
            self.clarify_client.create_bundle(name=name, media_url=media_url,
                                              metadata=metadata, external_id=external_id)
        self.sync_stats['created'] += 1

    def _update_metadata_for_video(self, metadata_href, video):
        '''
        Update the metadata for the video if video has been updated in Brightcove since the bundle
        metadata was last updated.
        '''
        current_metadata = self.clarify_client.get_metadata(metadata_href)
        cur_data = current_metadata.get('data')

        if cur_data.get('updated_at') != video['updated_at']:
            self.log('Updating metadata for video {0}'.format(video['id']))
            if not self.dry_run:
                metadata = self._metadata_from_video(video)
                self.clarify_client.update_metadata(metadata_href, metadata=metadata)
            self.sync_stats['updated'] += 1

    def sync_bundles(self, delete_bundles=True, update_metadata=True, confirm_delete_fun=None, dry_run=False):

        self.dry_run = dry_run
        self.sync_stats.clear()

        if dry_run:
            self.log('-----------------------------------')
            self.log('DRY RUN - not modifying any bundles')
            self.log('-----------------------------------')

        self.log('Fetching bundles...')
        bundle_map = self._load_bundle_map()
        self.log('Fetching videos...')
        videos = self.bc_client.get_all_videos()

        self.log('Checking {0} videos...'.format(len(videos)))
        for video in videos:
            vid = video['id']
            bundle = bundle_map.get(vid)

            if bundle is None:
                # Create a bundle for the video
                self._create_bundle_for_video(video)

            elif update_metadata:
                # Update the metadata in the bundle for the video
                self._update_metadata_for_video(get_link_href(bundle, 'clarify:metadata'), video)

        if delete_bundles:
            self.log('Checking deleted videos...')
            # Delete bundles for missing videos
            existing_vids = set([x['id'] for x in videos])
            existing_bundles = set(bundle_map.keys())
            missing_bundles = existing_bundles - existing_vids
            if len(missing_bundles):
                for vid in missing_bundles:
                    bundle = bundle_map.get(vid)
                    if dry_run or confirm_delete_fun is None or \
                            confirm_delete_fun(bundle['name'], bundle['external_id']):
                        self.log('Delete bundle for video {0}'.format(bundle['external_id']))
                        if not dry_run:
                            self.clarify_client.delete_bundle(get_link_href(bundle, 'self'))
                        self.sync_stats['deleted'] += 1

        self.sync_stats['count'] = len(bundle_map) + self.sync_stats['created'] - self.sync_stats['deleted']
        self.log('done.')



import requests
import math
import time
import base64
import json
import sys
import os

# -------------------------------------------------
# Based on a Brightcove sample test script.
# - ported to Python 3
# - Made into a class
# - Added API access functions
# - Added error handling
# -------------------------------------------------

# This is a python script to test the CMS API.
# To use this script, edit the configuration file brightcove_oauth.json
# with your brightcove account ID, and a client ID and client secret for
# an Oauth credential that has CMS API - Videos Read permission.
# You can find instructions on how to generate Oauth credentials
# http://docs.brightcove.com/en/video-cloud/cms-api/getting-started/quick-start-cms.html

# This script demonstrates how to refresh the access token
# in handling 401 - Unauthorized errors from the CMS API
# Because the Oauth tokens have a 300 second time to live,
# The refresh logic to handle 401 errors will be a normal part of runtime behavior.

# Note that the client_id and client_secret secure the access to the CMS API
# Therefore, it is not advisable to expose them to browsers. These are meant for
# server to server communication to obtain an access token.

# The access token can be exposed to the browser. Its limited permissions and expiry
# time make limit the duration and scope of its usage should it be observed in network
# traffic or obtained from a browser.


def add_param_if_not_none(params, param, val):
    if val is not None:
        params[param] = val


class AuthError(Exception):
    def __init__(self):
        super().__init__('Brightcove Auth Error')


class TooManyRequestsError(Exception):
    def __init__(self):
        super().__init__('Brightcove Too Many Requests')


class BrightcoveAPIClient:
    CMS_Server = "cms.api.brightcove.com"
    Ingest_Server = "ingest.api.brightcove.com"

    def __init__(self, creds_file='brightcove_oauth.json'):
        self.credentials = self._load_secret(creds_file)
        self.current_token = None

    def _load_secret(self, creds_file):
        '''read the oauth secrets and account ID from a credentials configuration file'''
        try:
            with open(creds_file) as fp:
                creds = json.load(fp)
            return creds
        except Exception as e:
            sys.stderr.write("Error loading oauth secret from local file called '{0}'\n".format(creds_file))
            sys.stderr.write("\tThere should be a local OAuth credentials file \n")
            sys.stderr.write("\twhich has contents like this:\n")
            sys.stderr.write("""
                {
                "account_id": "1234567890001",
                "client_id": "30ff0909-0909-33d3-ae88-c9887777a7b7",
                "client_secret": "mzKKjZZyeW5YgsdfBD37c5730g397agU35-Dsgeox6-73giehbt0996nQ"
                }

              """)
            sys.stderr.write("\n")
            raise e

    # get the oauth 2.0 token
    def _get_auth_token(self, creds):

        url = "https://oauth.brightcove.com/v3/access_token"

        params = {
            "grant_type": "client_credentials"
        }

        client = creds["client_id"]
        client_secret = creds["client_secret"]

        authString = base64.encodebytes('{0}:{1}'.format(client, client_secret).encode('utf-8')) \
            .decode('ascii').replace('\n', '')

        headersMap = {
            "Content-Type": "application/x-www-form-urlencoded",
            "Authorization": "Basic " + authString
        }

        response = requests.post(url, params=params, headers=headersMap)

        if response.status_code == 200:
            result = response.json()
            return result["access_token"]

    # call CMS API
    def _try_request(self, token, account, server, method, requestUrl, params=None, data=None):

        requestUrl = 'https://' + server + requestUrl

        headersMap = {
            "Authorization": "Bearer " + token
        }

        if data is not None and isinstance(data, dict):
            headersMap["Content-Type"] = "application/json"
            data = json.dumps(data)

        response = requests.request(method, requestUrl, params=params, data=data, headers=headersMap)
        result = response.json()

        if response.status_code >= 200 and response.status_code <= 299:
            return result
        elif response.status_code == 401:
            # if we get a 401 it is most likely because the token is expired.
            raise AuthError()
        elif response.status_code == 429:
            raise TooManyRequestsError()
        elif response.status_code == 500 and len(result) > 0 and result[0].get('error_code') == 'TIMEOUT':
            raise TooManyRequestsError()
        else:
            sys.stderr.write("Error: " + requestUrl + "\n")
            sys.stderr.write(response.text)
            sys.stderr.write("\n")
            raise Exception('API_CALL_ERROR' + " error " + str(response.status_code))

    def _make_request(self, server, method, urlSubPath, params=None, data=None):

        if (not self.current_token):
            self.current_token = self._get_auth_token(self.credentials)

        account = self.credentials["account"]

        requestUrl = "/v1/accounts/" + account + urlSubPath

        while self.current_token:
            try:
                return self._try_request(self.current_token, account, server, method, requestUrl, params, data)
            except TooManyRequestsError:
                # Sleep and try again...
                time.sleep(1)
            except AuthError:
                # handle an auth error by re-fetching a auth token again
                self.current_token = self._get_auth_token(self.credentials)

        raise AuthError()

    def get_video_count(self, search_q=None):
        '''Return the number of videos in the account'''
        if search_q is not None:
            params = {'q': search_q}
        else:
            params = None

        url = "/counts/videos"
        result = self._make_request(self.CMS_Server, 'GET', url, params=params)
        return result['count']

    def get_videos(self, limit=None, offset=None, search_q=None, sort=None):

        if limit is not None or offset is not None or search_q is not None or sort is not None:
            params = {}
            add_param_if_not_none(params, 'limit', limit)
            add_param_if_not_none(params, 'offset', offset)
            add_param_if_not_none(params, 'q', search_q)
            add_param_if_not_none(params, 'sort', sort)
        else:
            params = None

        url = '/videos'
        result = self._make_request(self.CMS_Server, 'GET', url, params=params)
        return result

    def get_all_videos(self, search_q=None):
        '''
        Gets all the videos in an account by automatically paginating through getVideos().
        WARNING: Use with caution if you have thousands of videos!
        WARNING: If deletes are being done during this iteration, the list may be missing videos.
        '''
        page_size = 10
        current_page = 0
        total_count = self.get_video_count(search_q=search_q)
        total_page = math.ceil(total_count / page_size)
        # We sort by 'created_at' so any newly added videos will be on the last page.
        # We don't currently handle deletes during the iteration which could cause videos to be missed.
        sort = 'created_at'
        result = []
        while current_page < total_page:
            page_result = self.get_videos(limit=page_size, offset=current_page * page_size,
                                          search_q=search_q, sort=sort)
            result += page_result
            current_page += 1
        return result

    def get_video_sources(self, bcVideoId):
        url = '/videos/' + bcVideoId + '/sources'
        result = self._make_request(self.CMS_Server, 'GET', url)
        return result

    def post_video(self, videoUrl, name=None, ingestMedia=True):
        '''Post and optionally ingest media from the specified URL'''
        if name is None:
            name = os.path.basename(videoUrl)

        url = '/videos'
        data = {'name': name}
        new_video = self._make_request(self.CMS_Server, 'POST', url, data=data)
        if ingestMedia:
            self.ingest_video(new_video['id'], videoUrl)
        return new_video

    def ingest_video(self, bcVideoId, videoUrl):
        '''Returns something like {'id': 'e99451a5-1a09-4258-8e73-798c9b58f126'}'''

        ingestData = {
            "master": {"url": videoUrl},
            "profile": "balanced-nextgen-player"  # "balanced-high-definition"
        }
        url = '/videos/' + bcVideoId + '/ingest-requests'
        result = self._make_request(self.Ingest_Server, 'POST', url, data=ingestData)
        return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys


try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist upload')
    sys.exit()

readme = open('README.rst').read()
history = open('HISTORY.rst').read().replace('.. :changelog:', '')

setup(
    name='clarify_brightcove_sync',
    version='0.0.1',
    description='Commandline tool to sync a Brightcove video library to Clarify.',
    long_description=readme + '\n\n' + history,
    author='Ivo Rothschild',
    author_email='ivo@clarify.io',
    url='https://github.com/Clarify/clarify_brightcove_sync',
    packages=[
        'clarify_brightcove_sync',
    ],
    scripts=['bin/clarify_brightcove_sync'],
    package_dir={'clarify_brightcove_sync':
                 'clarify_brightcove_sync'},
    include_package_data=True,
    install_requires=[
        'requests',
        'clarify_python'
    ],
    license="MIT",
    zip_safe=False,
    keywords='clarify_brightcove_sync clarify brightcove video audio search',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        "Programming Language :: Python :: 3",
    ],
    tests_require=[
        # httpretty 0.8.11 breaks due to being unable to find / open
        # a requirements file
        # httpretty 0.8.12 breaks due to having invalid python3 lambda
        # syntax
        'httpretty<=0.8.10'
    ],
    test_suite='tests',
)
